#!/usr/bin/tclsh
########################################################################
# Copyright (c) 2004 by Hartmut Schwarz, matut@matut-spielwerk.de      #
#                                                                      #
# This program is free software; you can redistribute it and/or        #
# modify it under the terms of the GNU General Public License          #
# as published by the Free Software Foundation; either version 2       #
# of the License, or (at your option) any later version.               #
#                                                                      #
# This program is distributed in the hope that it will be useful,      #
# but WITHOUT ANY WARRANTY; without even the implied warranty of       #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         #
# GNU General Public License for more details.                         #
#                                                                      #
# You should have received a copy of the GNU General Public            #
# License along with this program; if not, write to the Free           #
# Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,       #
# MA 02111-1307, USA.                                                  #
########################################################################

########################################################################
# a csound interface script adding some extras to the normal orchestra #
# and score file handling                                              #
########################################################################

########################################################################
# extra tools                                                          #
########################################################################
source paths.tcl

########################################################################
# configuration section                                                #
########################################################################
set sr 44100.0;
set kr 11025.0;

########################################################################
# function definition section                                          #
########################################################################

# load common functions
source various.tcl;
source wavinfo.tcl;

# print a minimum of help information
proc printhelp { name } {
 puts "$name : csound interface script for Clockwork-Music production";
 puts "usage : $name \[-ST|-MA|-IR\] dirtmp dirorc dirsco dirinf dirimp dirout track";
 puts "-ST     = simple stereo output of right and left signal";
 puts "-MA     = stereo output of main signal for left and right (i.e. mono)";
 puts "-IR     = mix IRC of main signal with left and right signal";
 puts "dirtmp  = directory where to put temporary files";
 puts "dirorc  = directory where to find csound score files";
 puts "dirsco  = directory where to find csound score files";
 puts "dirinf  = directory where to find extra score info";
 puts "dirimp  = directory where to find impulse response files (.wav)";
 puts "dirout  = directory where to put resulting sound files"; 
 puts "track   = track name";
 puts "NOTE: all directory names without terminating separator";
 puts "NOTE: IRC stands for Impulse Response Convolution";
}

########################################################################
# program execution section                                            #
########################################################################

# evaluate mixer mode
set mixmode [eval_mix_mode];
if { $mixmode != "-MA" && $mixmode != "-ST" && $mixmode != "-IR" } {
 printhelp $argv0;
 exit 1;
}
set argc [llength $argv];

# check input
if { $argc != 7 } {
 printhelp $argv0;
 exit 1;
}

# start a list of temporary files to be removed at the end
set tmpfiles [list];

# get input parameters
set dirtmp [file normalize [lindex $argv 0]];
set dirorc [file normalize [lindex $argv 1]];
set dirsco [file normalize [lindex $argv 2]];
set dirinf [file normalize [lindex $argv 3]];
set dirimp [file normalize [lindex $argv 4]];
set dirout [file normalize [lindex $argv 5]];
set track  [lindex $argv 6];

# set up some file path names
set filorc $dirorc/$track.orc
set filsco $dirsco/$track.sco
set filinf $dirinf/convinfo.$track.tcl
set filout $dirout/$track.wav
set definf $dirinf/default.convinfo.tcl

# for IRC load additional convolution info
if { $mixmode == "-IR" } {
 source $definf;
 if { [file isfile $filinf] } {
  source $filinf;
 }
 if { $impnam == "none" } {
  # set mixmode to simple stereo if no impulse response is specified
  set mixmode "-ST";  
 } else {
  set filimp $dirimp/$impnam.wav;
  set filcon $dirtmp/$impnam.con;
 }
}

if { $mixmode == "-ST" } {
# for simple stereo output just run csound with original ORC/SCO files
 set cmdlist [list $csound -W -d -o $filout $filorc $filsco |& $cat];
 puts "$cmdlist";
 if { [catch { eval [linsert $cmdlist 0 exec] } result] != 0 } {
  puts "\[ERROR\] \"$csound\" failed";
  puts "\[NOTE\] message received from \"$csound\":";
  puts "$result"
  exit 1;
 } elseif { $result != "" } {
  puts $result;
 }
} else {
# for mono or IRC output ORC/SCO files need some modifications
 # IRC wants some extra info first
 if { $mixmode == "-IR" } {
  # get csound convolution pregap delay
  set smpimp [evalnumsamples $filimp];
  set durimp [expr double($smpimp)/$sr];
  if { [expr 1.0/$kr] <= $durimp } {
   set delay [expr ceil($durimp*$kr)/$kr];
  } else {
   set delay [expr $durimp*ceil(1.0/($kr*$durimp))];
  }
  # do the impulse response signal preanalysis
  if { ![file exists $filcon] || [file mtime $filimp] > [file mtime $filcon] } {
   set cmdlist [list $csound -Ucvanal $filimp $filcon |& $cat];
   puts "$cmdlist";
   if { [catch { eval [linsert $cmdlist 0 exec] } result] != 0 } {
    puts "\[ERROR\] \"$csound\" failed";
    puts "\[NOTE\] message received from \"$csound\":";
    puts "$result"
    exit 1;
   } elseif { $result != "" } {
    puts $result;
   }
  }
 } else {
  set delay 0;
 }
 
 # write a modified (temporary) SCO file and
 # incidentely get new total play time for IRC 
 set tmpsco $dirtmp/csound.sco
 set insco  [open $filsco "r"];
 set outsco [open $tmpsco "w"];
 while { [gets $insco line] != -1 } {
  if { [regexp {^; --- i-Statements ---} $line] } {
   # add output mixer instrument play statement
   puts $outsco $line;
   puts $outsco "i 99 0.0 -1 ; play signal output mixer instrument\n";
   continue;
  } elseif { $mixmode == "-IR" && [regexp {^f0 [0-9]*\.[0-9]+ ; total time control} $line] } {
   # modify total play time for IRC
   scan $line "f0 %f" dursnd;
   set durout [expr $delay + $dursnd];
   if { $keepend == "yes" } {
    set durout [expr $durout + $durimp];
   }
   puts $outsco [format "f0 %f ; total time control" $durout];
   continue;
  }
  # anything else to be transfered without modification
  puts $outsco $line;
 }
 close $insco; 
 close $outsco; 
 lappend tmpfiles $tmpsco;

 # write a modified (temporary) ORC file
 set tmporc $dirtmp/csound.orc
 set inorc  [open $filorc "r"];
 set outorc [open $tmporc "w"];
 # first transfer original ORC with some modifications
 while { [gets $inorc line] != -1 } {
  if { [regexp {^; globals} $line] } {
   # add some global variables
   puts $outorc $line;
   puts $outorc "gamain  init 0 ; main (mono) signal";
   puts $outorc "galeft  init 0 ; left (stereo) signal";
   puts $outorc "garight init 0 ; right (stereo) signal";
   continue;
  } elseif { [regexp {^outs} $line] } {
   puts $outorc "gamain  = gamain  + amain  ;";
   puts $outorc "galeft  = galeft  + aleft  ;";
   puts $outorc "garight = garight + aright ;";
   continue;
  }
  # anything else to be transfered without modification
  puts $outorc $line;
 }
 # then append the output signal mixer instrument
 if { $mixmode == "-IR" } {
  # calculate fade in / fade out times  
  set dur1 [expr ($durout-$delay)*$intime/100.0];
  set dur3 [expr ($durout-$delay)*$outime/100.0]
  set dur2 [expr $durout-$delay-$dur1-$dur3];
  if { $dur2 < 0 } {
   set dur2 0.0;
  }
  # calculate resulting amplitude factors
  set amorig [expr $amcomm*$amorig];
  set amconv [expr $amcomm*$amconv];
  # IRC mixer instrument
  puts $outorc "\n; IRC output mixer instrument";
  puts $outorc "instr 99";
  puts $outorc "kenv linseg 0, $delay, 0, $dur1, 1, $dur2, 1, $dur3, 0";
  puts $outorc "korig = 1.0 - [expr 1.0 - $amorig]*kenv"; 
  puts $outorc "kconv = .5*$amconv*kenv";
  puts $outorc "a1l delay galeft, $delay";
  puts $outorc "a1r delay garight, $delay";
  puts $outorc "a2l convolve gamain, \"$impnam.con\", 1";
  puts $outorc "a2r convolve gamain, \"$impnam.con\", 2";
  puts $outorc "outs korig*a1l+kconv*a2l, korig*a1r+kconv*a2r";
  puts $outorc "gamain  = 0";
  puts $outorc "galeft  = 0";
  puts $outorc "garight = 0";
  puts $outorc "endin";
 } else {
  # Mono mixer instrument
  puts $outorc "\n; Mono output mixer instrument";
  puts $outorc "instr 99";
  puts $outorc "outs .5*gamain, .5*gamain";
  puts $outorc "gamain  = 0";
  puts $outorc "galeft  = 0";
  puts $outorc "garight = 0";
  puts $outorc "endin";
 }
 close $inorc; 
 close $outorc; 
 lappend tmpfiles $tmporc;

 # check if we need an intermediate sound output file
 if { $delay > 0 } {
  set filsnd $dirtmp/csound.wav
  lappend tmpfiles $filsnd;
 } else {
  set filsnd $filout;
 }

 # change to temporary directory (i.e. where csound may look for the .con file)
 cd $dirtmp;
 
 # run the csound job proper
 set cmdlist [list $csound -W -d -o $filsnd csound.orc csound.sco |& $cat];
 puts "$cmdlist";
 if { [catch { eval [linsert $cmdlist 0 exec] } result] != 0 } {
  puts "\[ERROR\] \"$csound\" failed";
  puts "\[NOTE\] message received from \"$csound\":";
  puts "$result"
  exit 1;
 } elseif { $result != "" } {
  puts $result;
 }

 # cut off convolution pregap delay
 if { $delay > 0 } {
  set start [expr int(floor($delay*$sr))];
  set cmdlist [list $sox $filsnd $filout trim ${start}s |& $cat];
  puts "$cmdlist";
  if { [catch { eval [linsert $cmdlist 0 exec] } result] != 0 } {
   puts "\[ERROR\] \"$sox\" failed";
   puts "\[NOTE\] message received from \"$sox\":";
   puts "$result"
   exit 1;
  } elseif { $result != "" } {
   puts $result;
  }
 }
}

# remove temporary files
foreach path $tmpfiles {
 file delete $path;
}
